/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.zookeeper.server.util;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import org.apache.zookeeper.common.Time;

/**
 * A class that provides simple interval-based rate limiting implementation.
 */
public class RateLimiter {
    private final int rate;
    private final long intervalInMs;
    private long lastTimeReset;
    private final AtomicInteger remained;

    public RateLimiter(final int rate, final long interval, final TimeUnit unit) {
        this.rate = rate;
        this.intervalInMs = unit.toMillis(interval);
        this.lastTimeReset = Time.currentElapsedTime();
        this.remained = new AtomicInteger(rate);
    }

    public boolean allow() {
        final long now = Time.currentElapsedTime();

        // reset the rate if interval passed
        if (now > lastTimeReset + intervalInMs) {
            remained.set(rate);
            lastTimeReset = now;
        }

        int value = remained.get();
        boolean allowed = false;

        // to handle race condition
        while (!allowed && value > 0) {
            allowed = remained.compareAndSet(value, value - 1);
            value = remained.get();
        }
        return allowed;
    }
}
